#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#define ASCII_TABLE     256
#define FILE_SIZE       10240

void usage()
{
    fprintf(stderr,
            "Syntax: generate-text [-f file_size] [-s seed] <-n fichier> [-a|-t|-w|-b|-r]\n"
            "   -n: nom du fichier a générer (peut être utilisé plusieurs fois)\n"
            "   -a: Table ASCII\n"
            "   -t: Texte\n"
            "   -w: texte qui se compresse bien\n"
            "   -b: data qui se compresse mal\n"
            "   -r: data avec fréquence au hasard (-s contrôle la graine)\n"
            "   -f: taille des fichiers générés (pour -w/-b/-r)\n"
          );
    exit(EXIT_SUCCESS);
}

void exit_if(int condition, const char *prefix) {
    if (condition) {
        perror(prefix);
        exit(EXIT_FAILURE);
    }
}

int generate_ascii_table(FILE *out)
{
    unsigned char data[ASCII_TABLE];

    if (out == NULL) return 0;
    for (int i = 0; i < ASCII_TABLE; i ++)
        data[i] = i;
    return fwrite(data, ASCII_TABLE, 1, out) > 0;
}

int generate_sample_text(FILE *out)
{
    if (out == NULL) return 0;
    return fprintf(out, "Auprès De Mon Arbre\nGeorges Brassens\n\n") > 0
        && fprintf(out,
            "J'ai plaqué mon chêne\n"
            "Comme un saligaud,\n"
            "Mon copain le chêne,\n"
            "Mon alter ego,\n"
            "On était du même bois\n"
            "Un peu rustique, un peu brut,\n"
            "Dont on fait n'importe quoi\n"
            "Sauf, naturell'ment, les flûtes...\n"
            "J'ai maint'nant des frênes,\n"
            "Des arbres de Judée,\n"
            "Tous de bonne graine,\n"
            "De haute futaie...\n"
            "Mais, toi, tu manque' à l'appel,\n"
            "Ma vieill' branche de campagne,\n"
            "Mon seul arbre de Noël,\n"
            "Mon mât de cocagne !\n\n"
            "Auprès de mon arbre,\n"
            "Je vivais heureux,\n"
            "J'aurais jamais dû m'éloigner de mon arbre...\n"
            "Auprès de mon arbre,\n"
            "Je vivais heureux,\n"
            "J'aurais jamais dû le quitter des yeux...\n"
           ) > 0;
}

int last_index_under(int value, int length, int array[])
{
    for (int i = 0; i < length; i ++)
        if (array[i] > value)
            return i;
    return -1;
}

int generate_text_with_probability(FILE* out, int text_len, int frequency[])
{
    int table[ASCII_TABLE];
    table[0] = frequency[0];
    for (int i = 1; i < ASCII_TABLE; i++)
        table[i] = table[i - 1] + frequency[i];
    int max = table[ASCII_TABLE - 1];

    while (text_len-- > 0) {
        int c = last_index_under(rand()%max, ASCII_TABLE, table);
        if (fputc(c, out) == EOF)
            return 0;
    }
    return 1;
}

int generate_text_compress_well(FILE* out, int file_size)
{
    int table[ASCII_TABLE] = {};

    for (int i = 32; i < 127; i ++)
        table[i] = rand() % 100 + 1;

    return generate_text_with_probability(out, file_size, table);
}

int generate_compress_bad(FILE* out, int file_size)
{
    int table[ASCII_TABLE];
    for (int i = 0; i < ASCII_TABLE; i ++)
        table[i] = 1;

    return generate_text_with_probability(out, file_size, table);
}

int generate_compress_random(FILE* out, int file_size, int seed)
{
    srand(seed);

    int table[ASCII_TABLE];
    for (int i = 0; i < ASCII_TABLE; i ++)
        table[i] = rand() % 256;

    return generate_text_with_probability(out, file_size, table);
}

int main(int argc, char *argv[])
{
    int seed = time(NULL);
    int file_size = FILE_SIZE;
    FILE *current_file = stdout;

    int ch;
    while ((ch = getopt(argc, argv, "n:atf:wbrs:")) != -1) {
        switch (ch) {
            case 'n':
                if (current_file) fclose(current_file);
                exit_if(fopen(optarg, "w") == NULL, optarg);
                break;
            case 'a':
                exit_if(!generate_ascii_table(current_file), "ascii table");
                break;
            case 't':
                exit_if(!generate_sample_text(current_file), "sample text");
                break;
            case 'f':
                file_size = atoi(optarg);
                break;
            case 'w':
                exit_if(!generate_text_compress_well(current_file, file_size), "compress well");
                break;
            case 'b':
                exit_if(!generate_compress_bad(current_file, file_size), "compress bad");
                break;
            case 's':
                seed = atoi(optarg);
                break;
            case 'r':
                exit_if(!generate_compress_random(current_file, file_size, seed), "random prob");
                break;
            case '?':
            default:
                usage();
        }
    }

    return EXIT_SUCCESS;
}
